"""
A module for updating properties (title, abstract, PI and coI) of a provided project.
"""
import argparse as ap
import logging
import os
import sys
import warnings
from typing import List

from pymygdala import LogHandler, SendNRAOEvent
from s_code_project_updater import Telescope
from schema.model import Author, Project
from schema.pstmodel import Person, UserAuthentication
from sqlalchemy import exc as sa_exc, asc, desc
from support.capo import get_my_capo_config
from support.logging import get_console_logger, LOG_MESSAGE_FORMATTER

from schema import ArchiveDBSession, ExecutionBlock
from ._version import ___version___ as version
from .project_fetcher import ArchiveProjectFetcher

_APPLICATION_NAME = 's_code_project_updater'
_LOG = get_console_logger(_APPLICATION_NAME, logging.DEBUG)
_MISSING_PROFILE = """ERROR: unknown 'profile', provide the -P argument or set the CAPO_PROFILE
 environment variable."""
_DESCRIPTION = """A tool for updating the investigators, title and/or abstract of a project."""
_EPILOG = """Return values:
0: everything worked,
1: error with capo configuration
2: Error with input parameters
3: Project not found
4: Investigator not found
5: Update failed"""


class ScodeProjectUpdater:
    """
    A class to bundle the operations involved with updating a project in the archive.
    """
    def __init__(self, **kwargs):

        """
        Build an instance of the class with the cmd line parser, the capo_config, a context
        manager for the archive and pst, and do some rudamentary tests of the input arguments
        so we fail early.
        :param kwargs: the command line arguments or namespace with the arguments to the parser
        """
        self._make_parser()
        try:
            _LOG.warning('parsing....')
            self.args = self.parser.parse_args(**kwargs)
            _LOG.warning('parsed')

        except Exception as exc:
            _LOG.error(f'parser threw {exc}')
            self.exit_with_error('Capo profile and project code are '
                                 'required', 2)

        # at a minimum, Capo profile and project code are required
        if not self.args.profile or not self.args.project:
            if not self.args.profile and not self.args.project:
                self.exit_with_error('Capo profile and project code are '
                                     'required', 2)
            if not self.args.profile:
                self.exit_with_error('Capo profile not specified', 2)
            if not self.args.project:
                self.exit_with_error('project code not specified', 2)

        args_dict = self.args.__dict__

        if args_dict['dry']:
            self.is_dry = args_dict['dry']
        else:
            self.is_dry = False

        if not args_dict['investigators'] and not args_dict['title'] and not args_dict['abstract']:
            self.set_minimum_properties_from_args(args_dict)
            # return

        self.project_code = args_dict['project']
        self.stored_project = None
        _LOG.debug(f'{self.args}')

        try:
            self.capo_config = get_my_capo_config(profile=self.args.profile)
        except Exception as exc:
            self.exit_with_error(f'Capo configuration error: {exc}', 1)
        try:
            self.archive_context = ArchiveDBSession('SDM', profile=self.capo_config.profile)
            self.pst_context = ArchiveDBSession('PST', profile=self.capo_config.profile)
        except KeyError as k_ex:
            self.exit_with_error(f'An error occurred while creating a db context: {k_ex}', 1)

        if self.args.investigators and \
                [inv for inv in self.args.investigators if self.args.investigators.count(inv) > 1]:
            self.exit_with_error('You appear to be trying to add an investigator more than once, '
                                 'which could cause issues with the presentation of investigators '
                                 'in the archive.  There should be only one PI and any number of '
                                 'unique CoIs on a project.', 2)


    def scode_project_from_args(self, args: ap.Namespace):
        ''' turn the command-line arguments into a project '''

        ns_dict = args.__dict__
        project_code = ns_dict['project']
        title = ns_dict['title']
        abstract = ns_dict['abstract']
        investigators = ns_dict['investigators']

        new_project = ArchiveProject(project_code=project_code,
                                     title=title,
                                     abstract=abstract,
                                     author_pst_ids=investigators)
        return new_project


    def set_minimum_properties_from_args(self, args):
        ''' basic info needed for a fetch '''
        self.project_code = args['project']
        self.profile = args['profile']
        self.is_dry = True
        self.fetch_only = True

    def _make_parser(self):

        r""" Build a command line parser for this app. """
        result = ap.ArgumentParser(description=_DESCRIPTION.format(version),
                                   formatter_class=ap.RawTextHelpFormatter,
                                   epilog=_EPILOG)
        result.add_argument('-C', '--project', action='store',
                            help='project_code to update')
        result.add_argument('-P', '--profile', action='store',
                            help='profile name to use, e.g. test, production')
        result.add_argument('-T', '--title', action='store',
                            help='a quoted string for the new title for the project')
        result.add_argument('-A', '--abstract', action='store',
                            help='a quoted string for the new abstract for the project')
        result.add_argument('-I', '--investigators', action='store', type=int, nargs='+',
                            help='a PST ID, or list of PST IDs, of investigators for the project, '
                                 'as an unquoted integer or space seperated integer list.  The '
                                 'first ID in the list will be added as the PI and all subsequenct '
                                 'IDs will be added as CoIs.')
        result.add_argument('-d', '--dry', action='store_true',
                            help='perform a dry run, going through the motions, but not committing '
                                 'changes and not performing a re-index of the project.  This may '
                                 'be useful because it will print the current state of the project '
                                 'and what the project would look like after the changes.')
        self.parser = result

    def exit_with_error(self, msg, code):
        """
        On discovering we have an unresolvable condition the prevents us from proceeding with the
        update of this project, print an error and exit with the code provided.
        :param msg: an error message to the user
        :param code: the exit code to accompany the error message
        :return: None
        """
        self.error_message = msg
        self.code = code
        _LOG.error(f'error message received: {self.error_message}; code = {self.code}')
        _LOG.error(msg)
        self.parser.print_help()
        sys.exit(code)

    def get_stored_project(self):
        """
        Return the project specified by the input arguments, if it exists.
        :return: the first Project we found with the project code passed in
        """
        project = self.archive_context.session.query(Project) \
            .filter(Project.project_code == self.args.project) \
            .first()
        return project

    def get_pst_users(self, investigators):
        """
        Get the Person(s) associated with the investigators (a list of PST person ID(s)).
        :return: the Person(s) mapped to the person ID(s) passed in
        """
        users = self.pst_context.session.query(Person.person_id, Person.firstName,
                                               Person.lastName, UserAuthentication.personName)\
            .join(UserAuthentication,
                  Person.personAuthentication_id == UserAuthentication.userAuthentication_id)\
            .filter(Person.person_id.in_(investigators)).all()
        return users

    def get_projects_current_investigators(self):
        """
        Get a list of investigators associated with this project, ordered such that the PI(s)
        is/are the first element(s) of the list
        :return: a list of investigators associated with the project code passed in, ordered with
        the PI(s) first
        """
        investigators_list = self.archive_context.session.query(Author) \
            .filter(Author.project_code == self.args.project) \
            .order_by(desc(Author.is_pi), asc(Author.pst_person_id)) \
            .all()
        return investigators_list

    def clear_projects_current_investigators(self):
        """
        Clear the investigators currently attached to this project
        :return: None
        """
        investigators_list = self.archive_context.session.query(Author) \
            .filter(Author.project_code == self.args.project) \
            .all()
        for inv in investigators_list:
            self.archive_context.session.delete(inv)

    def set_new_project_investigators(self, new_investigators):
        """
        Add a list of new project investigators
        :param new_investigators: a list of investigators with PST personName, firstName, lastName
        and person_id
        :return: None
        """

        # if any of the new investigators already exists,
        # use the old author_id rather than making a new author
        is_pi = True
        num_expected = len(new_investigators)
        num_changed  = 0
        # Loop through our cmd line investigator list and map them to investigator list passed in.
        # On the first pass through, we set the Author as the PI.
        for in_inv in self.args.investigators:
            for pst_user in new_investigators:
                if in_inv == int(pst_user.person_id):
                    # The Author table has the project as a foreign key, so we use the whole
                    # project here, rather than just a string, sqlalchemy wires the rest.  If you
                    # just give the table a project_code as a string, you will get an error.
                    # noinspection PyTypeChecker
                    auth = Author(author_id=None,
                                  project=self.stored_project,
                                  username=pst_user.personName,
                                  firstname=pst_user.firstName,
                                  lastname=pst_user.lastName,
                                  pst_person_id=str(pst_user.person_id),
                                  is_pi=is_pi)
                    self.archive_context.session.add(auth)
                    num_changed += 1
                    is_pi = False
                    break

        if num_changed < num_expected:
            _LOG.error(f'{num_changed} of {num_expected} investigators were NOT set')
            raise UpdateException('incomplete investigator update')

    def print_project(self):
        """
        Print the project's current investigators
        :return: None
        """
        output = self.get_project_info()
        [_LOG.info(line) for line in output]

    def get_project_info(self):
        output = []
        output.append(f'Title: {self.stored_project.title}')
        output.append(f'Abstract: {self.stored_project.abstract}')
        investigator_list = self.get_projects_current_investigators()

        # we want the PI's pst_person_id followed by the CoIs' pst_person_ids in numeric order
        pi = investigator_list[0]
        if pi.pst_person_id is not None:
            coi_pst_ids = [int(coi.pst_person_id) for coi in investigator_list[1:]]
            coi_pst_ids = sorted(coi_pst_ids)
            author_pst_ids = [int(pi.pst_person_id)]
            [author_pst_ids.append(id) for id in coi_pst_ids]
            authors_to_print = [str(id) for id in author_pst_ids]
            id_list = ' '.join(authors_to_print)
            output.append(f'Authors: {id_list}')

        return output

    def is_fetch_only(self):
        try:
            return self.fetch_only
        except AttributeError:
            return False

    def update_project(self) -> Project:
        '''
            Where the magic happens: update aspects of the project
            (including authors) according to the arguments passed in
        :return:
        '''
        fetcher = ArchiveProjectFetcher(self.args.profile)
        try:
            self.project = fetcher.fetch_project(self.project_code)
        except AttributeError:
            self.exit_with_error(f'project code "{self.project_code}" not '
                                 f'found', 3)
        if self.is_fetch_only():
            output = fetcher.build_project_info()
            try:
                [_LOG.info(line) for line in output]
            except TypeError:
                _LOG.error('Cannot display project info; is this an ALMA project?')
            return self.project

        """
        The main function responsible for updating the project.  It makes sure the project exists,
        and, if the user is updating the investigators, that they have valid PST mappings.  If there
        aren't errors with those two checks it clears the projects current archive authors and
        replaces them with the investigators found from the PST mapping to users.  And, of course,
        if the title and abstract are being updated, it adds those to the project
        :return: None
        """
        with warnings.catch_warnings(), self.archive_context, self.pst_context:
            # Suppress SQLAlchemy warnings
            warnings.simplefilter("ignore", category=sa_exc.SAWarning)

            # Get the project or fail
            self.stored_project = self.get_stored_project()
            if self.stored_project is None:
                self.exit_with_error('No project found for the project_code provided', 3)

            if self.is_alma():
                self.exit_with_error(f'{self.stored_project.project_code} '
                                 f'is an ALMA project; update not permitted', 2)

            if self.args.investigators:
                proposed_investigators = self.get_pst_users(self.args.investigators)
                if len(proposed_investigators) == 0 or \
                        not len(self.args.investigators) == len(proposed_investigators):
                    self.exit_with_error('One or more of the investigators you entered was not '
                                         'found in the PST.', 4)
                self.clear_projects_current_investigators()
                self.set_new_project_investigators(proposed_investigators)

            if self.args.title:
                self.stored_project.title = self.args.title
            if self.args.abstract:
                self.stored_project.abstract = self.args.abstract

            if not self.args.dry:
                if not self.is_fetch_only():
                    self.archive_context.session.commit()
                    _LOG.info('Changes committed')
            elif not self.is_fetch_only():
                _LOG.info('Successful dry run; this would have updated the project')

            self.print_project()
            return self.stored_project

    def is_alma(self):
        ''' is this an alma project? '''
        with warnings.catch_warnings():
            # Suppress SQLAlchemy warnings
            warnings.simplefilter("ignore", category=sa_exc.SAWarning)

            exec_block = self.archive_context.session.query(ExecutionBlock) \
                .filter(ExecutionBlock.project_code == self.project_code) \
                .filter(ExecutionBlock.telescope == Telescope.ALMA.value) \
                .first()
            return exec_block is not None



    def reindex_project(self):
        """
        If we are not performing a dry run, and have made it this far without error, then we
        re-index the project so the updates will show up in the profile-mapped archive.
        :return: None
        """
        if not self.args.dry and not self.is_fetch_only() \
                and '_TEST_PROJECT' not in self.project_code:
            _LOG.info(f'Re-indexing project {self.args.project} to make changes available....')
            # Set up a LogHandler to record the fact we just made a change to this project.
            # We're adding it here, instead of earlier, because nothing we log earlier should be
            # presented to anyone but the command line user and would only add useless clutter to
            # our system logging. We only really want the completed task to make a record in our
            # system.
            broadcast = LogHandler(profile=self.capo_config.profile, application=_APPLICATION_NAME)
            broadcast.setLevel(logging.DEBUG)
            broadcast.setFormatter(LOG_MESSAGE_FORMATTER)
            _LOG.addHandler(broadcast)
            _LOG.info(f'Project {self.args.project} has been updated.')

            event = {'logData': {'project_code': self.args.project,
                                 'title_updated': self.args.title is not None,
                                 'abstract_updated': self.args.abstract is not None,
                                 'investigators_updated': self.args.investigators is not None,
                                 'ingestion_type': 'evla_sdm'
                                 },
                     'message': 's-code project updated',
                     'request': 're-index please'}
            SendNRAOEvent(profile=self.capo_config.profile, application=_APPLICATION_NAME) \
                .send(routing_key='ingestion-complete.metadata', event=event)
            sys.exit(0)

class ArchiveProject:

    """
    A class to encapsulate the attributes of a project as stored in the archive
    for the purpose of updating and getting current information about the project.
    """

    def __init__(self, project_code: str, title: str, abstract: str, author_pst_ids: List, ):
        """
        Represents the state of a project.
        :param project_code:
        :param title:
        :param abstract:
        :param author_pst_ids:
        """
        self.project_code = project_code
        self.title = title
        self.abstract = abstract
        self.investigators = author_pst_ids

        options = []
        options.append('-C')
        options.append('--project')
        options.append('-P')
        options.append('--profile')
        options.append('-T')
        options.append('--title')
        options.append('-A')
        options.append('--abstract')
        options.append('-I')
        options.append('--investigators')
        self.options = options

    def make_args(self, is_dry):
        args = []
        if is_dry:
            args.append('-d')

        args.append('-C')
        args.append(self.project_code)
        args.append('-P')
        self.profile = os.environ['CAPO_PROFILE']
        args.append(self.profile)
        args.append('-T')
        args.append(self.title)
        args.append('-A')
        args.append(self.abstract)

        if self.investigators:
            args.append('-I')
            for pst_id in self.investigators:
                args.append(str(pst_id))

        return args

    def is_arg(self, arg):
        return arg in self.options

    def add_parameter(self, new_project, key, value):
        if '-C' == key or '--project' == key:
            new_project.project_code = value
        elif '-T' == key or '--title' == key:
            new_project.title = value
        elif '-A' == key or '--abstract' == key:
            new_project.abstract = value
        elif '-P' == key or '--profile' == key:
            self.profile = value

    def add_investigators(self, new_project, args, start_position):
        value = args[start_position]

        while value not in self.options:
            new_project.investigators.append(value)

        return new_project.investigators


class UpdateArgParser:
    ''' Command-line argument parser for ScodeProjectUpdater '''
    def __init__(self):
        self.parser = self._make_parser()

    def _make_parser(self):
        parser = ap.ArgumentParser(description=_DESCRIPTION.format(version),
                                   formatter_class=ap.RawTextHelpFormatter,
                                   epilog=_EPILOG)
        parser.add_argument('-C', '--project', action='store',
                            help='project_code to update')
        parser.add_argument('-P', '--profile', action='store',
                            help='profile name to use, e.g. test, production')
        parser.add_argument('-T', '--title', action='store',
                            help='a quoted string for the new title for the project')
        parser.add_argument('-A', '--abstract', action='store',
                            help='a quoted string for the new abstract for the project')
        parser.add_argument('-I', '--investigators', action='store', type=int, nargs='+',
                            help='a PST ID, or list of PST IDs, of investigators for the project, '
                                 'as an unquoted integer or space seperated integer list.  The '
                                 'first ID in the list will be added as the PI and all subsequenct '
                                 'IDs will be added as CoIs.')
        parser.add_argument('-d', '--dry', action='store_true',
                            help='perform a dry run, going through the motions, but not committing '
                                 'changes and not performing a re-index of the project.  This may '
                                 'be useful because it will print the current state of the project '
                                 'and what the project would look like after the changes.')
        return parser

    def parse_args(self, **kwargs):
        return self.parser.parse_args(kwargs)


class UpdateException(Exception):
    ''' throw this if there is trouble during the update '''
    pass

def main(**kwargs):
    """
    The script's main entry point.
    :param kwargs: the command line arguments
    :return: None
    """
    updater = ScodeProjectUpdater(**kwargs)
    updater.update_project()
    # reindex only if there's been an update
    if not updater.is_dry:
        updater.reindex_project()


if __name__ == '__main__':
    main()
