import logging
import os
import subprocess
import unittest
import warnings

import pytest
from s_code_project_updater.commands import UpdateException, ScodeProjectUpdater
from schema import create_session
from schema.model import Project
from schema.pstmodel import Session
from sqlalchemy import exc as sa_exc
from support.logging import get_console_logger

from .test_projects import get_author_pst_ids, ScodeTestProject, \
    ScienceTestProject, AlmaTestProject

_LOG = get_console_logger("scode_project_updater_tests", logging.DEBUG)
_UPDATE_COMMAND = 'update_sproj'
PROFILE = 'local'

class UpdaterTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        os.environ['CAPO_PROFILE'] = PROFILE
        cls.return_values = build_updater_return_values()

    @classmethod
    def setUp(cls) -> None:
        cls.initialize_test_data(cls)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.remove_test_data(cls)

    def test_dry_run_does_not_update(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        session = create_session('SDM')
        return_code = None
        try:
            new_title = 'this is the new title'
            self.assertNotEqual(fake_project.title, new_title,
                                f'new title should be {new_title}; got '
                                f'{fake_project.title}')
            args = [
                '-C', project_code,
                '-P', PROFILE,
                '-T', new_title,
                '--dry'
                ]
            try:
                return_code = CommandLineUpdaterLauncher(args).run()
            except Exception as exc:
                text = self.return_values[return_code] if return_code else ''
                pytest.fail(f'{exc} {text}')

            if not return_code:
                updated = self.get_project_from_db(session, project_code)
                # nothing should have been updated
                self.assertEqual(fake_project.title, updated.title,
                                 f'expecting same title, but before is '
                                 f'{fake_project.title} and after is {updated.title}')
                self.assertEqual(fake_project.abstract, updated.abstract,
                                 f'expecting same abstract, but before is '
                                 f'{fake_project.abstract} and updated is {updated.abstract}')
                self.assertEqual(len(fake_project.authors),
                                     len(updated.authors),
                                 f'expecting same number of authors, '
                                 f'but before has {len(fake_project.authors)} '
                                 f'and after has {len(updated.authors)}')
            else:
                pytest.fail(f'unexpected failure; return code ={return_code}')

        except Exception as exc:
            pytest.fail(f'{project_code}: {exc}')
        finally:
            session.close()

    def test_project_code_only_fetches(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        args = [
            '-C', project_code,
            '-P', PROFILE,
        ]
        return_code = None
        try:
            return_code = CommandLineUpdaterLauncher(args).run()
            if not return_code:
                session = create_session('SDM')
                try:
                    updated = self.get_project_from_db(session, project_code)
                    self.assertEqual(fake_project.title, updated.title,
                                     f'expecting same title, but before is '
                                     f'{fake_project.title} and after is {updated.title}')
                    self.assertEqual(fake_project.abstract, updated.abstract,
                                     f'expecting same abstract, but before is '
                                     f'{fake_project.abstract} and updated is {updated.abstract}')
                    self.assertEqual(len(fake_project.authors),
                                     len(updated.authors),
                                     f'expecting same number of authors, '
                                     f'but before has {len(fake_project.authors)} '
                                     f'and after has {len(updated.authors)}')
                    count = 0
                    for orig_author in fake_project.authors:
                        for author in updated.authors:
                            if author.username == orig_author.username:
                                count += 1
                                break
                    self.assertEqual(len(fake_project.authors), count,
                                     f'before and after projects should have '
                                     f'same authors')
                finally:
                    session.close()
        except Exception as exc:
            text = self.return_values[return_code] if return_code else ''
            pytest.fail(f'{exc} {text}')

    def test_updates_abstract_only(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        session = create_session('SDM')
        new_abstract = "Well, here's another nice mess you've gotten us into, Ollie"
        self.assertNotEqual(fake_project.abstract, new_abstract,
                            f'expecting new abstract {new_abstract} but got {fake_project.abstract}')
        args = [
            '-C', project_code,
            '-P', PROFILE,
            '-A', new_abstract,
        ]
        return_code = None
        try:
            try:
                return_code = CommandLineUpdaterLauncher(args).run()
            except subprocess.TimeoutExpired as exp:
                raise UpdateException(exp)
            except Exception as exc:
                text = self.return_values[return_code] if return_code else ''
                pytest.fail(f'{exc} {text}')

            if not return_code:
                updated = self.get_project_from_db(session, project_code)
                # only abstract should have been updated;
                # all else should be same
                self.assertEqual(fake_project.title, updated.title,
                                 f'expecting same title, but before is '
                                 f'{fake_project.title} and after is {updated.title}')
                self.assertEqual(new_abstract, updated.abstract,
                                 f'expecting same abstract, but before is '
                                 f'{fake_project.abstract} and updated is {updated.abstract}')
                self.assertEqual(len(fake_project.authors),
                                 len(updated.authors))
            else:
                raise UpdateException()

        except Exception as exc:
            pytest.fail(f'{project_code}: {exc}')
        finally:
            session.close()

    def test_updates_abstract_and_title(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        session = create_session('SDM')
        new_abstract = "I think you ought to know I'm feeling very depressed"
        new_title = 'A Survey of the Mattresses of Sqornshellous Zeta'
        self.assertNotEqual(fake_project.abstract, new_abstract,
                            f'expecting new abstract {new_abstract}, '
                            f'but abstract was not changed from {fake_project.abstract}')
        self.assertNotEqual(fake_project.title, new_title,
                            f'expecting new title {new_title}, '
                            f'but abstract was not changed from {fake_project.title}')
        args = [
            '-C', project_code,
            '-P', PROFILE,
            '-A', new_abstract,
            '-T', new_title,
        ]
        return_code = None
        try:
            try:
                return_code = CommandLineUpdaterLauncher(args).run()
            except subprocess.TimeoutExpired as exp:
                raise UpdateException(exp)
            except Exception as exc:
                text = self.return_values[return_code] if return_code else ''
                pytest.fail(f'{exc} {text}')

            if not return_code:
                updated = self.get_project_from_db(session, project_code)
                # abstract and title should have been updated;
                # all else should be same
                self.assertEqual(new_title, updated.title,
                                 'title should  not have changed')
                self.assertEqual(new_abstract, updated.abstract,
                                 'abstract should not have changed')
                self.assertEqual(len(fake_project.authors),
                                 len(updated.authors),
                                 'authors should not have changed')
            else:
                raise UpdateException()

        except Exception as exc:
            pytest.fail(f'{project_code}: {exc}')
        finally:
            session.close()

    def test_adds_new_abstract_deletes_author(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        new_project = Project(project_code=project_code,
                              title=fake_project.title,
                              abstract=fake_project.abstract)
        new_abstract = "First there is a mountain, then there is no " \
                        "mountain, then there is"
        self.assertNotEqual(new_abstract, fake_project.abstract)
        new_project.abstract = new_abstract
        original_authors = fake_project.authors.copy()
        self.assertEqual(4, len(original_authors),
                         'expected 4 authors before update')
        last_author = original_authors[3]
        new_authors = original_authors[:3]
        self.assertEqual(len(original_authors) - 1, len(new_authors),
                         f'expecting {len(original_authors) - 1} new authors, '
                         f'but there are {len(new_authors)}')
        new_project.authors = new_authors
        args = [
            '-C', project_code,
            '-P', PROFILE,
            '-A', new_abstract,
            '-I',
        ]
        for id in get_author_pst_ids(new_project):
            args.append(str(id))

        return_code = None
        session = create_session('SDM')
        try:
            try:
                return_code = CommandLineUpdaterLauncher(args).run()
            except subprocess.TimeoutExpired as exp:
                raise UpdateException(exp)
            except Exception as exc:
                text = self.return_values[return_code] if return_code else ''
                pytest.fail(f'{exc} {text}')

            self.assertEqual(0, return_code, f'command should have succeeded '
                                             f'but return code was {return_code}')

            updated = self.get_project_from_db(session, project_code)
            # last author should have been removed and the abstract changed;
            # title should remain same
            self.assertNotEqual(fake_project.abstract, updated.abstract,
                                'abstract should have changed')
            self.assertEqual(fake_project.title, updated.title,
                             'title should not have changed')
            expected = len(original_authors) - 1
            actual = len(updated.authors)
            self.assertEqual(expected, actual,
                             'one author should have been removed')
            authors_updated = last_author in updated.authors
            self.assertFalse(authors_updated, 'THIS IS THE MESSAGE')
            count = 0
            for orig_author in original_authors[:3]:
                for new_author in updated.authors:
                    if new_author.username == orig_author.username:
                        count += 1
                        break
            self.assertEqual(len(new_authors), count,
                             f'expected {len(new_authors)} authors in '
                             f'updated project; there were {count}')

        except Exception as exc:
            pytest.fail(f'{project_code}: {exc}')

        finally:
            session.close()

    def test_output_is_as_expected(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        args = [
            '-C', project_code,
            '-P', PROFILE,
        ]

        runner = CommandLineUpdaterLauncher(args)
        return_code = runner.run()
        if return_code:
            text = self.return_values[return_code]
            pytest.fail(text)

        stdout = runner.stdout
        self.assertIsNotNone(stdout, 'program output is expected')
        self.assertTrue('Title: ' + fake_project.title in stdout,
                        'title should be in output')
        self.assertTrue('Abstract: ' + fake_project.abstract in stdout,
                        'abstract should be in output')
        pst_ids = [str(id) for id in get_author_pst_ids(fake_project)]
        pst_id_str = ' '.join(pst_ids)
        self.assertTrue('Authors: ' + pst_id_str in stdout,
                        f'output should have PST IDs {pst_ids}')

    def test_copes_with_single_pi(self):
        project = ScodeTestProject().project
        args = ['-P', PROFILE, '-C', project.project_code, '-I', '4686']
        return_code = CommandLineUpdaterLauncher(args=args).run()
        self.assertEqual(0, return_code,
                         'update to single author should succeed')

    def test_alma_project_is_rejected(self):
        project_code = '2018.A.00062.S'
        args = ['-P', PROFILE, '-C', project_code,
                '-T', 'Physics at High Angular Resolution in Nearby Galaxies: '
                      'The Local Galaxy Inventory Continued']

        with pytest.raises(SystemExit) as exc:
            ScodeProjectUpdater(args=args).update_project()
            self.assertEqual(2, exc.code, 'ALMA project should be rejected')

    def test_errors_return_expected_codes(self):
        # minimum required arguments -- profile & project -- omitted
        return_code = CommandLineUpdaterLauncher([]).run()
        self.assertEqual(return_code, 2,
                         'expected return code 2 for no args')

        project_code = ScodeTestProject().project.project_code

        # update failure
        result = FailingUpdater().update_project()
        self.assertIsInstance(result, SystemExit)
        self.assertEqual(5, result.code,
                         'expecting return code 5 for update failure')

        # profile not specified
        args = ['-C', project_code,]
        return_code = CommandLineUpdaterLauncher(args).run()
        self.assertEqual(return_code, 2,
                         'expecting return code 2 when profile not specified')

        # project code not specified
        args = ['-P', PROFILE]
        self.assertEqual(CommandLineUpdaterLauncher(args).run(), 2,
                         'expecting return code 2 when project not specified')

        # profile value missing
        args = ['-P', '-C', project_code]
        return_code = CommandLineUpdaterLauncher(args).run()
        self.assertEqual(return_code, 2,
                         'expecting return code 2 for missing profile')

        # project code missing
        args = ['-P', PROFILE, '-C']
        self.assertEqual(CommandLineUpdaterLauncher(args).run(), 2,
                         'expecting return code 2 for missing project code')

        # bad project code
        args = ['-P', PROFILE, '-C', 'bogus']
        self.assertEqual(CommandLineUpdaterLauncher(args).run(), 3,
                         'expecting return code 3 for invalid project code')

        # bad profile
        args = ['-P', 'not_a_profile', '-C', project_code]
        self.assertEqual(CommandLineUpdaterLauncher(args).run(), 1,
                         'expecting return code 1 for invalid Capo profile')

        # missing title as last argument
        args = ['-P', PROFILE, '-C', project_code, '-T']
        self.assertEqual(CommandLineUpdaterLauncher(args).run(), 2,
                         'expecting return code 2 for missing title')

        # missing title as first argument
        args = [ '-T', '-P', PROFILE, '-C', project_code,]
        self.assertEqual(CommandLineUpdaterLauncher(args).run(), 2,
                         'expecting return code 2 for missing title')

        # nonexistent investigator
        args = ['-P', PROFILE, '-C', project_code, '-I', '-22']
        self.assertEqual(CommandLineUpdaterLauncher(args).run(), 4,
                         'expecting return code 4 for invalid investigator')


    ### UTILITIES ###

    def initialize_test_data(self):
        session = create_session('SDM')
        num_commits = num_found = 0
        try:
            with warnings.catch_warnings():
                # Suppress SQLAlchemy warnings
                warnings.simplefilter("ignore", category=sa_exc.SAWarning)

                fake_projects = [ScodeTestProject().project,
                                 ScienceTestProject().project,
                                 AlmaTestProject().project]
                try:
                    for fake_project in fake_projects:
                        project_code = fake_project.project_code
                        existing = session.query(Project) \
                            .filter(Project.project_code ==
                                    project_code) \
                            .first()
                        if existing is not None:
                            num_found += 1
                            session.delete(existing)
                            session.commit()
                        session.add(fake_project)
                        session.commit()
                        num_commits += 1

                    if num_commits < num_found:
                        pytest.fail(f'{num_found} fake projects were found '
                                    f'and deleted, but {num_commits} were '
                                    f'added and committed')
                except Exception as exc:
                    pytest.fail(f'{exc}')
        finally:
            session.close()

    def remove_test_data(self):
        session = create_session('SDM')
        try:
            with warnings.catch_warnings():
                # Suppress SQLAlchemy warnings
                warnings.simplefilter("ignore", category=sa_exc.SAWarning)

                fake_projects = [ScodeTestProject().project,
                                 ScienceTestProject().project,
                                 AlmaTestProject().project]
                for fake_project in fake_projects:
                    project_code = fake_project.project_code
                    existing = session.query(Project) \
                        .filter(Project.project_code ==
                                project_code) \
                        .first()
                    if existing is not None:
                        session.delete(existing)
                        session.commit()
        except Exception as exc:
            pytest.fail(f'{exc}')
        finally:
            session.close()

    def get_project_from_db(self, session: Session, project_code: str):
        with warnings.catch_warnings():
            # Suppress SQLAlchemy warnings
            warnings.simplefilter("ignore", category=sa_exc.SAWarning)

            return session.query(Project) \
                .filter(Project.project_code == project_code) \
                .first()

class FailingUpdaterHelper:
    # def __init__(self, **kwargs):
    #     pass

    @pytest.fixture()
    def update_project(self):
        return SystemExit(5)

class FailingUpdater:
    def __init__(self):
        self.helper = FailingUpdaterHelper()

    def update_project(self):
        return SystemExit(5)


class CommandLineUpdaterLauncher:

    def __init__(self, args: list):
        self.args = [_UPDATE_COMMAND]
        for arg in args:
            self.args.append(str(arg))
        _LOG.info(f'{self.args}')

    def run(self):
        ''' launch updater from command line
            @:returns directory listing
        '''

        args = self.args
        try:
            proc = subprocess.run(args,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT,
                                       timeout=60,
                                       check=False,
                                       bufsize=1,
                                       universal_newlines=True)
            self.stdout = proc.stdout
            return proc.returncode
        except Exception as exc:
            _LOG.error(f'{exc}')
            if not isinstance(exc, subprocess.TimeoutExpired):
                return exc.returncode
            else:
                raise

def build_updater_return_values():
    return {
        1: 'error with capo configuration',
        2: 'error with input parameters',
        3: 'project not found',
        4: 'investigator not found',
        5: 'update failed',
    }

if __name__ == '__main__':
    unittest.main()
